# Python API Extractor to connect to Coupa source system and pull the requisite tables for UiPath Process Mining P2P Discovery Accelerator

# Import the necessary tables
import requests
import json
import pandas as pd
from sqlalchemy import create_engine
from dotenv import dotenv_values
import sys


# Connection to Coupa API

# Save requisite information for API connection

# Load environment variables from .env file
try:
    env_vars = dotenv_values('extractor_settings.env')

    coupa_url = env_vars['url']
    client_id = env_vars['client_id']
    client_secret = env_vars['client_secret']
    scopes = env_vars['scopes']

    print(f"Successfully loaded environment variables from 'extractor_settings.env' file.")

except FileNotFoundError:
    print("Failed to find 'extractor_settings.env' file.")
    sys.exit(1)

# Set headers

headers = {
    "Accept" : "application/json",
    "Content-Type" : "application/x-www-form-urlencoded"

}

# Set payload

payload = {
    "grant_type": "client_credentials",
    "client_id" : client_id,
    "client_secret" : client_secret,
    "scope": scopes
}

# Send request to get access token

response = requests.post(f"{coupa_url}/oauth2/token", headers=headers, data=payload)


if response.status_code == 200:
    data = json.loads(response.content)
    access_token = data["access_token"]
    headers["Authorization"] = f"Bearer {access_token}"
    print("Connection to Coupa API established. Access token retrieved.")
else:
    print(f"Failed to establish connection to Coupa API. Response code: {response.status_code}")
    sys.exit(1)


# Save the variables that are going to be used. This includes start_date, end_date, and the directory for csv files to be written to.


start_date = env_vars['start_date']
end_date = env_vars['end_date']
csv_directory = env_vars['csv_directory']
database = env_vars['target_database']
csv_choice = env_vars['csv_choice']

# Snowflake connection for SQLAlchemy
if not csv_choice:
    if database == 'snowflake':
        #Import requisite library for snowflake
        from snowflake.sqlalchemy import URL
        engine = create_engine(URL(
            user= env_vars['snowflake_user']
            , password= env_vars['snowflake_password']
            , account= env_vars['snowflake_account']
            , database= env_vars['snowflake_database']
            , schema= env_vars['snowflake_base_schema']))
        connection = engine.connect()

    elif database == 'sqlserver':
        #Import requisite library for snowflake
        import pyodbc
        #Create engine
        sql_driver= env_vars['sql_driver']
        sql_dsn= env_vars['sql_dsn']
        sql_server= env_vars['sql_server']
        sql_user= env_vars['sql_user']
        sql_password= env_vars['sql_password']
        sql_database= env_vars['sql_database']
        sql_schema= env_vars['sql_schema']

        # Create the SQLAlchemy engine
        engine = create_engine(f'mssql+pyodbc://{sql_user}:{sql_password}@{sql_server}/{sql_database}?driver=ODBC+Driver+17+for+SQL+Server')

        connection = engine.connect()

    else:
        raise ValueError("Invalid database choice. Please select either snowflake or sqlserver.")
        sys.exit(1)

    print("Connection to Database successful.")

# Fields for the input tables

approval_fields = json.dumps([
    "id",
    "created-at",
    "status",
    "approver-id",
    "position",
    "approval-date",
    "note",
    "approvable-id",
    "approvable-type"
])

invoice_line_fields = json.dumps([
    "id",
    "invoice-number",
    {"invoice_lines": [
        "id",
        "created-at",
        "description",
        "line-num",
        "order-header-num",
        "order-line-id",
        "order-line-num",
        "total",
        "status",
        "type",
        "quantity",
        "accounting-total",
        {"uom": ["name"]}
    ]}
])

invoices_fields = json.dumps([
    "id",
    "created-at",
    "invoice-date",
    "delivery-number",
    "delivery-date",
    "status",
    "discount-due-date",
    "net-due-date",
    "discount-percent",
    "payment-method",
    "invoice-number",
    "paid",
    "payment-date",
    "payment-notes",
    "payment-order-reference",
    "payment-order-number",
    "clearance-document",
    {
        "created_by": [
            "id",
            "fullname"
        ]
    },
    {
        "payment_term": [
            "code",
            "days-for-net-payment",
            "days-for-discount-payment",
            "discount-rate"
        ]
    },
    {
        "pay_invoice": [
            "id",
            "created-at",
            "status",
            "payable-amount",
            "paid-amount",
            "paid-date"
        ]
    },
    {
        "custom_fields": [
            "document-date",
            "posting-date"
        ]
    },
    {
        "uom": [
            "name"
        ]
    },
    {
        "currency": [
            "code"
        ]
    }
])

purchase_order_line_fields = json.dumps([
    "id",
    "line-num",
    "created-at",
    "need-by-date",
    "description",
    "requisition-line-id",
    "status",
    "total",
    "accounting-total",
    "quantity",
    "order-header-id",
    "order-header-number",
    {
        "commodity": [
            "name"
        ]
    },
    {
        "account": [
            "name"
        ]
    },
    {
        "created_by": [
            "id"
        ]
    },
    {
        "requester": [
            "id"
        ]
    },
    {
        "uom": [
            "name"
        ]
    },
    {
        "currency": [
            "code"
        ]
    }
])

purchase_order_revisions_fields = json.dumps([
    "id",
    {
        "revision_record": [
            "revision",
            "id",
            "created-at",
            "created-by",
            "exported"
        ]
    },
    {
        "order_lines": [
            "id",
            {
                "accounting-total": ["from", "to"]
            },
            {
                "status": ["from", "to"]
            },
            {
                "version": ["from", "to"]
            }
        ]
    }
])

purchase_orders_fields = json.dumps([
    "id",
    "created-at",
    "po-number",
    "status",
    "internal-revision",
    "change-type",
    {
        "requisition_header": [
            "id"
        ]
    },
    {
        "created_by": [
            "id",
            "fullname"
        ]
    },
    {
        "supplier": [
            "id",
            "name"
        ]
    },
    {
        "ship_to_address": [
            "business-group-name"
        ]
    }
])

receiving_transactions_fields = json.dumps([
    "id",
    "created-at",
    "status",
    "quantity",
    "comments",
    {
        "order_line": [
            "id"
        ]
    },
    {
        "created_by": [
            "id"
        ]
    }
])

requisition_line_fields = json.dumps([
    "id",
    "status",
    "requisition-id",
    "line-num",
    "created-at",
    "description",
    "receipt-required",
    "source-type",
    "quantity",
    "total",
    {
        "commodity": [
            "name",
            "category"
        ]
    },
    {
        "uom": [
            "name"
        ]
    },
    {
        "created_by": [
            "id",
            "fullname"
        ]
    },
    {
        "currency": [
            "code"
        ]
    }
])

requisition_fields = json.dumps([
    "id",
    "created-at",
    "status",
    {
        "requested_by": [
            "id",
            "fullname"
        ]
    },
    {
        "requisition_lines": [
            "id"
        ]
    }
])

suppliers_fields = json.dumps([
    "id",
    "name",
    "display-name",
    "one-time-supplier",
    {
        "primary_address": [
            "state",
            {
                "country": [
                    "code",
                    "name"
                ]
            }
        ]
    }
])

users_fields = json.dumps([
    "id",
    "fullname",
    {
        "department": [
            "name"
        ]
    }
])

# Creation of functions to extract the necessary tables and data

# General function for extracting tables
def table_extractor(table_name, start_date, end_date, fields):
    data = []
    offset = 0
    limit = 50

    while True:
        response = requests.get(f"{coupa_url}/api/{table_name}?offset={offset}&created_at[gt_or_eq]={start_date}&created_at[lt]={end_date}&fields={fields}", headers=headers)
        response_data = response.json()
        data.extend(response_data)

        if len(response_data) < limit:
            break  # Break the loop if the number of returned records is less than the limit

        offset += limit

    df = pd.json_normalize(data)

    # Replace hyphens and periods with underscores in column names
    df.columns = df.columns.str.replace('-', '_')
    df.columns = df.columns.str.replace('.', '_')

    # Write table to database or csv
    if csv_choice:
        df.to_csv(f'{csv_directory}/{table_name}.csv', sep='\t', index=False)
    else:
        df.to_sql(f'{table_name}', con=engine, index=False, if_exists='replace')
        connection.close()
        engine.dispose()

    return print(f"{table_name} successfully written to directory. {table_name} table created. Total records fetched: {len(df)}")



# Function for pulling the correct requisition lines
def requisitions_table_extractor(start_date,end_date):
    data = []
    offset = 0
    limit = 50

    while True:
        response = requests.get(f"{coupa_url}/api/requisitions?offset={offset}&created_at[gt_or_eq]={start_date}&created_at[lt]={end_date}&fields={requisition_fields}", headers=headers)
        response_data = response.json()
        data.extend(response_data)

        if len(response_data) < limit:
            break  # Break the loop if the number of returned records is less than the limit

        offset += limit
    
    # Parsing through all requisition and requisition line information
    requisition_lines_list = []
    for requisition in data:
        for requisition_line in requisition['requisition-lines']:
            requisition_line_dict = {}
            requisition_line_dict['requisition-id'] = requisition['id']
            requisition_line_dict['created-at'] = requisition['created-at']
            requisition_line_dict['status'] = requisition['status']
            requisition_line_dict['requested-by-id'] = requisition['requested-by']['id']
            requisition_line_dict['requested-by-fullname'] = requisition['requested-by']['fullname']
            requisition_line_dict['requisition-line-id'] = requisition_line['id']
            requisition_lines_list.append(requisition_line_dict)
    requisitions_df = pd.DataFrame(requisition_lines_list)

    # Replace hyphens and periods with underscores in column names
    requisitions_df.columns = requisitions_df.columns.str.replace('-', '_')
    requisitions_df.columns = requisitions_df.columns.str.replace('.', '_')

    # Creating a list of Users for filtering Users Table
    global pr_users
    pr_users = requisitions_df['requested_by_id'].unique()
    pr_users = [int(item) for item in pr_users]
    pr_users = json.dumps(pr_users)

    #Use the below to write the table to database or csv
    if csv_choice:
        requisitions_df.to_csv(f'{csv_directory}/requisitions.csv', sep='\t', index=False)
    else:
        requisitions_df.to_sql("requisitions", con=engine, index=False, if_exists='replace') #make sure index is False
        connection.close()
        engine.dispose()
    
    return print(f"requisitions successfully written to directory. requisitions table created. Total records fetched: {len(requisitions_df)}")

# Function for pulling the correct purchase orders and also creating a list of supplier ID's
def purchase_orders_table_extractor(start_date, end_date, fields):
    data = []
    offset = 0
    limit = 50

    while True:
        response = requests.get(f"{coupa_url}/api/purchase_orders?offset={offset}&created_at[gt_or_eq]={start_date}&created_at[lt]={end_date}&fields={fields}", headers=headers)
        response_data = response.json()
        data.extend(response_data)

        if len(response_data) < limit:
            break  # Break the loop if the number of returned records is less than the limit

        offset += limit

    df = pd.json_normalize(data)

    # Replace hyphens and periods with underscores in column names
    df.columns = df.columns.str.replace('-', '_')
    df.columns = df.columns.str.replace('.', '_')

    # Creating a list of Suppliers for filtering Supplier Table
    global po_suppliers
    po_suppliers = df["supplier_id"].unique()
    po_suppliers = [int(item) for item in po_suppliers]
    po_suppliers = json.dumps(po_suppliers)

    # Writing to either database or CSV
    if csv_choice:
        df.to_csv(f'{csv_directory}/purchase_orders_source.csv', sep='\t', index=False)
    else:
        df.to_sql(name = f"purchase_orders".lower(), con=engine, index=False, if_exists='replace')
        connection.close()
        engine.dispose()

    return print(f"purchase orders successfully written to directory. purchase_orders table created. Total records fetched: {len(df)}")


# Function for extracting invoices table
def invoices_table_extractor(start_date, end_date):
    data = []
    offset = 0
    limit = 50

    while True:
        response = requests.get(f"{coupa_url}/api/invoices?offset={offset}&created_at[gt_or_eq]={start_date}&created_at[lt]={end_date}&fields={invoices_fields}", headers=headers)
        response_data = response.json()
        data.extend(response_data)

        if len(response_data) < limit:
            break  # Break the loop if the number of returned records is less than the limit

        offset += limit

    # Handle missing payment_term attributes
    for invoice in data:
        payment_term = invoice.get('payment_term')
        if payment_term is None:
            payment_term = {
                "code": None,
                "days-for-net-payment": None,
                "days-for-discount-payment": None,
                "discount-rate": None
            }
            invoice['payment-term'] = payment_term

    df = pd.json_normalize(data)

    # Replace hyphens and periods with underscores in column names
    df.columns = df.columns.str.replace('-', '_')
    df.columns = df.columns.str.replace('.', '_')

    # Write table to database or csv
    if csv_choice:
        df.to_csv(f'{csv_directory}/invoices_source.csv', sep='\t', index=False)
    else:
        df.to_sql(f'invoices', con=engine, index=False, if_exists='replace')
        connection.close()
        engine.dispose()

    return print(f"invoices successfully written to directory. invoices table created. Total records fetched: {len(df)}")


# Function to pull invoice line information
def invoice_lines_extractor(start_date,end_date):
    data = []
    offset = 0
    limit = 50

    while True:
        response = requests.get(f"https://uipath-stage.coupahost.com/api/invoices?offset={offset}&created_at[gt_or_eq]={start_date}&created_at[lt]={end_date}&fields={invoice_line_fields}", headers=headers)
        response_data = response.json()
        data.extend(response_data)

        if len(response_data) < limit:
            break  # Break the loop if the number of returned records is less than the limit

        offset += limit
    
    invoice_lines_list = []
    for invoice in data:
        for invoice_line in invoice['invoice-lines']:
            invoice_line_dict = {}
            invoice_line_dict['invoice-id'] = invoice['id']
            invoice_line_dict['invoice-number'] = invoice['invoice-number']
            invoice_line_dict['id'] = invoice_line['id']
            invoice_line_dict['created-at'] = invoice_line['created-at']
            invoice_line_dict['description'] = invoice_line['description']
            invoice_line_dict['line-num'] = invoice_line['line-num']
            invoice_line_dict['order-header-num'] = invoice_line['order-header-num']
            invoice_line_dict['order-line-id'] = invoice_line['order-line-id']
            invoice_line_dict['order-line-num'] = invoice_line['order-line-num']
            invoice_line_dict['quantity'] = invoice_line['quantity']
            invoice_line_dict['status'] = invoice_line['status']
            invoice_line_dict['total'] = invoice_line['total']
            invoice_line_dict['type'] = invoice_line['type']
            invoice_line_dict['accounting-total'] = invoice_line['accounting-total']
            #invoice_line_dict['uom-name'] = invoice_line['uom']['name']
            
            uom = invoice_line.get('uom')
            if uom:
                invoice_line_dict['uom-name'] = uom['name']
            else:
                invoice_line_dict['uom-name'] = None
            invoice_lines_list.append(invoice_line_dict)
    invoice_lines_df = pd.DataFrame(invoice_lines_list)

    # Replace hyphens and periods with underscores in column names
    invoice_lines_df.columns = invoice_lines_df.columns.str.replace('-', '_')
    invoice_lines_df.columns = invoice_lines_df.columns.str.replace('.', '_')

    #Use the below to write the table to database or csv
    if csv_choice:
        invoice_lines_df.to_csv(f'{csv_directory}/invoice_lines.csv', sep='\t', index=False)
    else:
        invoice_lines_df.to_sql("invoice_lines", con=engine, index=False, if_exists='replace') #make sure index is False, Snowflake doesnt accept indexes
        connection.close()
        engine.dispose()

    return print(f"invoice lines successfully written to directory. invoice_lines table created. Total records fetched: {len(invoice_lines_df)}")


# Function to connect to purchase order revisions API endpoint and write tables to csv and database
def purchase_order_revisions_table_extractor(start_date, end_date):
    # Extracting the required attributes from the JSON response
    data = []
    offset = 0
    limit = 50

    while True:
        response = requests.get(f"{coupa_url}/api/purchase_order_revisions?offset={offset}&created_at[gt_or_eq]={start_date}&created_at[lt]={end_date}&fields={purchase_order_revisions_fields}", headers=headers)
        response_data = response.json()
        data.extend(response_data)

        if len(response_data) < limit:
            break  # Break the loop if the number of returned records is less than the limit

        offset += limit

    revision_list = []
    for item in data:
        row = {
            'id': item['id'],
            'revision': item['revision-record']['revision'],
            'revision_id': item['revision-record']['id'],
            'created_at': item['revision-record']['created-at'],
            'created_by': item['revision-record']['created-by'],
            'exported': item['revision-record']['exported'],
        }
        order_lines = item.get('order-lines', [])
        if order_lines:
            order_line = order_lines[0]
            row['order_line_id'] = str(order_line['id'])
            accounting_total = order_line.get('accounting-total', {})
            row['accounting_total_from'] = accounting_total.get('from')
            row['accounting_total_to'] = accounting_total.get('to')
            
            # Adding the order line status and version attributes
            status = order_line.get('status', {})
            row['status_from'] = status.get('from')
            row['status_to'] = status.get('to')
            
            version = order_line.get('version', {})
            row['version_from'] = version.get('from')
            row['version_to'] = version.get('to')
        else:
            row['order_line_id'] = None
            row['accounting_total_from'] = None
            row['accounting_total_to'] = None
            row['status_from'] = None
            row['status_to'] = None
            row['version_from'] = None
            row['version_to'] = None
        revision_list.append(row)
    
    # Create a DataFrame from the extracted data
    df = pd.DataFrame(revision_list)

    #Use the below to write the table to databse or csv
    if csv_choice:
        df.to_csv(f'{csv_directory}/purchase_order_revisions.csv', sep='\t', index=False)
    else:
        df.to_sql("purchase_order_revisions", con=engine, index=False, if_exists='replace') #make sure index is False, Snowflake doesnt accept indexes
        connection.close()
        engine.dispose()

    return print(f"purchase order revisions successfully written to directory. purchase_order_revisions table created. Total records fetched: {len(df)}")


# Function for pulling the master tables
def master_table_extractor(table_name, fields, id_list):
    data = []
    offset = 0
    limit = 50

    while True:
        response = requests.get(f"{coupa_url}/api/{table_name}?offset={offset}&id[in]={id_list}&fields={fields}", headers=headers)
        response_data = response.json()
        data.extend(response_data)

        if len(response_data) < limit:
            break  # Break the loop if the number of returned records is less than the limit

        offset += limit

    df = pd.json_normalize(data)

    # Replace hyphens and periods with underscores in column names
    df.columns = df.columns.str.replace('-', '_')
    df.columns = df.columns.str.replace('.', '_')


    # Write tables to database or csv
    if csv_choice:
        df.to_csv(f'{csv_directory}/{table_name}.csv', sep='\t', index=False)
    else:
        df.to_sql(f'{table_name}', con=engine, index=False, if_exists='replace')
        connection.close()
        engine.dispose()

    return print(f"{table_name} successfully written to directory. {table_name} table created. Total records fetched: {len(df)}")


# Function for pulling the supplier master table
def supplier_table_extractor(fields, id_list):
    data = []
    offset = 0
    limit = 50

    while True:
        response = requests.get(f"{coupa_url}/api/suppliers?offset={offset}&id[in]={id_list}&fields={fields}", headers=headers)
        response_data = response.json()
        data.extend(response_data)

        if len(response_data) < limit:
            break  # Break the loop if the number of returned records is less than the limit

        offset += limit

    # # Handle missing country attributes
    for supplier in data:
        country = supplier.get('country')
        if country is None:
            country = {
                "code": None,
                "name": None
            }
            supplier['primary_address_country'] = country

    df = pd.json_normalize(data)

    # Replace hyphens and periods with underscores in column names
    df.columns = df.columns.str.replace('-', '_')
    df.columns = df.columns.str.replace('.', '_')

    # Remove duplicate columns
    df = df.loc[:, ~df.columns.duplicated()]

    # Write tables to database or csv
    if csv_choice:
        df.to_csv(f'{csv_directory}/suppliers.csv', sep='\t', index=False)
    else:
        df.to_sql('suppliers', con=engine, index=False, if_exists='replace')
        connection.close()
        engine.dispose()

    return print(f"suppliers successfully written to directory. suppliers table created. Total records fetched: {len(df)}")


# Runs the invoice lines function for invoice lines
invoice_lines_extractor(start_date, end_date)

# Runs the above to pull Requisitions_raw
requisitions_table_extractor(start_date, end_date)

# Runs the table_extractor function for requisition lines
table_extractor('requisition_lines', start_date, end_date, requisition_line_fields)

# Runs the purchase_order_extractor function for purchase orders
purchase_orders_table_extractor(start_date, end_date, purchase_orders_fields)

# Runs the table_extractor function for purchase order lines
table_extractor('purchase_order_lines', start_date, end_date, purchase_order_line_fields)

# Runs the table_extractor function for receiving transactions
table_extractor('receiving_transactions', start_date, end_date, receiving_transactions_fields)

# Runs the table_extractor function for approvals
table_extractor('approvals', start_date, end_date, approval_fields)

# Runs invoice extractor function for invoices
invoices_table_extractor(start_date, end_date)

# Runs the purchase order revisions function for purchase order revisions
purchase_order_revisions_table_extractor(start_date, end_date)

# Runs the supplier_table_extractor function. Pulls filtered table with just supplier ID's found in Purchase Orders
supplier_table_extractor(suppliers_fields, po_suppliers)

# Runs the master_table_extractor function for users. Pulls whole table. May take a while
master_table_extractor('users', users_fields, pr_users)


print("Python Extractor has run successfully. All requisite tables have been extracted.")
sys.exit(0)
